﻿using System;
using System.Threading.Tasks;
using Un4seen.Bass;
using Windows.Media.Core;
using Windows.Storage;
using Windows.Storage.FileProperties;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Media.MediaProperties;

namespace BackgroundPlayer
{
  class BassMediaSource : IMediaSource
  {
    MediaStreamSource m_MediaStreamSource = null;
    BASS_CHANNELINFO m_Info = null;
    string m_FilePath = string.Empty;
    ulong m_FileSize = 0;
    double m_SecondsPosition = 0;
    int m_BassHandle = 0;

    public BassMediaSource(string filePath)
    {
      if (string.IsNullOrEmpty(filePath))
        throw new ArgumentNullException("filePath");
      m_FilePath = filePath;
    }

    public MediaStreamSource GetMediaSource()
    {
      return m_MediaStreamSource;
    }

    public async Task InitializeAsync()
    {
      StorageFile sFile = await StorageFile.GetFileFromPathAsync(m_FilePath);
      BasicProperties prop = await sFile.GetBasicPropertiesAsync();
      m_FileSize = prop.Size;

      m_BassHandle = Bass.BASS_StreamCreateFile(m_FilePath, 0, (long)m_FileSize, BASSFlag.BASS_STREAM_DECODE);
      if (m_BassHandle == 0) {
        BASSError err = Bass.BASS_ErrorGetCode();
        System.Diagnostics.Debug.WriteLine("InitializeAsync error {0}", err);
      }

      m_Info = Bass.BASS_ChannelGetInfo(m_BassHandle);
      if (m_Info == null) {
        BASSError err = Bass.BASS_ErrorGetCode();
        System.Diagnostics.Debug.WriteLine("InitializeAsync error {0}", err);
      }
      long len = Bass.BASS_ChannelGetLength(m_BassHandle, BASSMode.BASS_POS_BYTES);
      double secs = Bass.BASS_ChannelBytes2Seconds(m_BassHandle, len);
      uint bits = 16;
      if (m_Info.Is32bit)
        bits = 32;
      else if (m_Info.Is8bit)
        bits = 8;

      AudioEncodingProperties pcmprops = AudioEncodingProperties.CreatePcm((uint)m_Info.freq, (uint)m_Info.chans, bits);
      m_MediaStreamSource = new MediaStreamSource(new AudioStreamDescriptor(pcmprops));
      m_MediaStreamSource.CanSeek = true;
      m_MediaStreamSource.BufferTime = TimeSpan.Zero;
      m_MediaStreamSource.Duration = TimeSpan.FromSeconds(secs);
      m_MediaStreamSource.Closed += mss_Closed;
      m_MediaStreamSource.Starting += mss_Starting;
      m_MediaStreamSource.SampleRequested += mss_SampleRequested;
    } // InitializeAsync

    public static async Task<BassMediaSource> CreateAsync(string file)
    {
      BassMediaSource res = new BassMediaSource(file);
      await res.InitializeAsync();
      return res;
    } // CreateAsync

    #region MediaStreamSource 
    void mss_Starting(MediaStreamSource sender, MediaStreamSourceStartingEventArgs args)
    {
      var deferral = args.Request.GetDeferral();
      System.Diagnostics.Debug.WriteLine("mss_Starting {0}", args.Request.StartPosition);
      if (args.Request.StartPosition == null) {
        args.Request.SetActualStartPosition(TimeSpan.FromSeconds(0));
      } else if (args.Request.StartPosition != null) {
        var seconds = args.Request.StartPosition.Value.TotalSeconds;
        m_SecondsPosition = seconds;
        args.Request.SetActualStartPosition(TimeSpan.FromSeconds(this.m_SecondsPosition));
      }

      deferral.Complete();
    } // mss_Starting

    void mss_SampleRequested(MediaStreamSource sender, MediaStreamSourceSampleRequestedEventArgs args)
    {
      var deferral = args.Request.GetDeferral();
      System.Diagnostics.Debug.WriteLine("mss_SampleRequested");

      try {
        byte[] buffer = new byte[4096];
        int decoded = Bass.BASS_ChannelGetData(m_BassHandle, buffer, buffer.Length);
        if (decoded == -1) {
          BASSError err = Bass.BASS_ErrorGetCode();
          System.Diagnostics.Debug.WriteLine("mss_SampleRequested error {0}", err);
        } else {
          double secs = Bass.BASS_ChannelBytes2Seconds(m_BassHandle, decoded);

          MediaStreamSample sample = MediaStreamSample.CreateFromBuffer(buffer.AsBuffer(), TimeSpan.FromSeconds(m_SecondsPosition));
          sample.Duration = TimeSpan.FromSeconds(secs);
          m_SecondsPosition += sample.Duration.TotalSeconds;
          args.Request.Sample = sample;
        }
      } catch { }
      deferral.Complete();
    } // mss_SampleRequested

    void mss_Closed(MediaStreamSource sender, MediaStreamSourceClosedEventArgs args)
    {
      System.Diagnostics.Debug.WriteLine("mss_Closed {0}", args.Request.Reason);
      if (m_MediaStreamSource != null) {
        m_MediaStreamSource.Closed -= mss_Closed;
        m_MediaStreamSource.Starting -= mss_Starting;
        m_MediaStreamSource.SampleRequested -= mss_SampleRequested;
        m_MediaStreamSource = null;
      }

      if (m_BassHandle != 0) {
        Bass.BASS_StreamFree(m_BassHandle);
        m_BassHandle = 0;
      }
    } // mss_Closed
    #endregion
  }
}
